/*
 * kgraft_patch_bsc1090338
 *
 * Fix for CVE-2018-3665, bsc#1090338
 *
 *  Upstream commits:
 *  58122bf1d856 ("x86/fpu: Default eagerfpu=on on all CPUs")
 *  4ecd16ec7059 ("x86/fpu: Fix math emulation in eager fpu mode")
 *  6e6867093de3 ("x86/fpu: Fix eager-FPU handling on legacy FPU machines")
 *  814fb7bb7db5 ("x86/fpu: Don't let userspace set bogus xcomp_bv")
 *
 *  SLE12 commit:
 *  none yet
 *
 *  SLE12-SP1 commit
 *  none yet
 *
 *  SLE12-SP2 commit:
 *  73dfaa56b0cc56630790ca5e7dad422aa591cc16 (stable 4.4.90)
 *
 *  SLE12-SP3 commit:
 *  19373fd1583ccc3e579fd459f102fba7f8bdf98a
 *  4ca41b9382bb4c330347e4dc06d5512b2cd6ffac (stable 4.4.90)
 *
 *  Copyright (c) 2018 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#if IS_ENABLED(CONFIG_X86_64)

#define pr_fmt(fmt) "kgraft-patch: " fmt

#include <linux/kernel.h>
#include <linux/kallsyms.h>
#include <linux/sched.h>
#include <linux/tracepoint.h>
#include <linux/signal.h>
#include <linux/uaccess.h>
#include <linux/ptrace.h>
#include <asm/ptrace.h>
#include <asm/processor.h>
#include <asm/i387.h>
#include <asm/fpu-internal.h>
#include <asm/xsave.h>
#include <asm/sigframe.h>
#include <asm/syscall.h>
#include <trace/events/sched.h>
#include "patch_state.h"
#include "kgr_patch_bsc1090338.h"

#if IS_ENABLED(CONFIG_X86_32)
#error "Live patch supports only CONFIG_X86_32=n."
#endif

#if IS_ENABLED(CONFIG_MATH_EMULATION)
#error "Live patch supports only CONFIG_MATH_EMULATION=n."
#endif

#define BSC1090338_SUBPATCH_ID KGR_SUBPATCH_ID_BSC(1090338)


struct xsave_struct **kgr_init_xstate_buf;
static u64 *kgr_pcntxt_mask;
static unsigned int *kgr_mxcsr_feature_mask;
static struct task_struct * __percpu *kgr_fpu_owner_task;
static void (*kgr_flush_ptrace_hw_breakpoint)(struct task_struct *tsk);
static int (*kgr_ia32_setup_rt_frame)(int sig, struct ksignal *ksig,
				      compat_sigset_t *set,
				      struct pt_regs *regs);
static int (*kgr_ia32_setup_frame)(int sig, struct ksignal *ksig,
				   compat_sigset_t *set, struct pt_regs *regs);
static int (*kgr_save_xstate_sig)(void __user *buf, void __user *buf_fx,
				  int size);
static int (*kgr_copy_siginfo_to_user)(siginfo_t __user *to, siginfo_t *from);
static int (*kgr_setup_sigcontext)(struct sigcontext __user *sc,
				   void __user *fpstate, struct pt_regs *regs,
				   unsigned long mask);
static void (*kgr_signal_setup_done)(int failed, struct ksignal *ksig,
				     int stepping);
static int (*kgr_get_signal_to_deliver)(siginfo_t *info,
					struct k_sigaction *return_ka,
					struct pt_regs *regs, void *cookie);
static void (*kgr__set_current_blocked)(const sigset_t *newset);
static void (*kgr_convert_to_fxsr)(struct task_struct *tsk,
				   const struct user_i387_ia32_struct *env);
static void (*kgr_math_state_restore)(void);

static void (*kgr_synchronize_sched)(void);


static struct {
	char *name;
	char **addr;
} kgr_funcs[] = {
	{ "init_xstate_buf", (void *)&kgr_init_xstate_buf },
	{ "pcntxt_mask", (void *)&kgr_pcntxt_mask },
	{ "mxcsr_feature_mask", (void *)&kgr_mxcsr_feature_mask },
	{ "fpu_owner_task", (void *)&kgr_fpu_owner_task },
	{ "flush_ptrace_hw_breakpoint",
		(void *)&kgr_flush_ptrace_hw_breakpoint },
	{ "ia32_setup_rt_frame", (void *)&kgr_ia32_setup_rt_frame },
	{ "ia32_setup_frame", (void *)&kgr_ia32_setup_frame },
	{ "save_xstate_sig", (void *)&kgr_save_xstate_sig },
	{ "copy_siginfo_to_user", (void *)&kgr_copy_siginfo_to_user },
	{ "setup_sigcontext", (void *)&kgr_setup_sigcontext },
	{ "signal_setup_done", (void *)&kgr_signal_setup_done },
	{ "get_signal_to_deliver", (void *)&kgr_get_signal_to_deliver },
	{ "__set_current_blocked", (void *)&kgr__set_current_blocked },
	{ "convert_to_fxsr", (void *)&kgr_convert_to_fxsr },
	{ "math_state_restore", (void *)&kgr_math_state_restore },

	{ "synchronize_sched", (void *)&kgr_synchronize_sched },
};


/* from linux/tracepoint.h */
static inline void kgr_tracepoint_synchronize_unregister(void)
{
	kgr_synchronize_sched();
}


/* from arch/x86/include/asm/fpu-internal.h */
/* use the kallsyms-resolved kgr_fpu_owner_task() */
static inline void kgr__thread_clear_has_fpu(struct task_struct *tsk)
{
	tsk->thread.fpu.has_fpu = 0;
	this_cpu_write(*kgr_fpu_owner_task, NULL);
}

/* use the kallsyms-resolved kgr_fpu_owner_task() */
static inline void kgr__thread_set_has_fpu(struct task_struct *tsk)
{
	tsk->thread.fpu.has_fpu = 1;
	this_cpu_write(*kgr_fpu_owner_task, tsk);
}

static inline void kgr__thread_fpu_end(struct task_struct *tsk)
{
	kgr__thread_clear_has_fpu(tsk);
	if (!use_eager_fpu())
		stts();
}

static inline void kgr__thread_fpu_begin(struct task_struct *tsk)
{
	if (!static_cpu_has_safe(X86_FEATURE_EAGER_FPU))
		clts();
	kgr__thread_set_has_fpu(tsk);
}

static inline void kgr__drop_fpu(struct task_struct *tsk)
{
	if (__thread_has_fpu(tsk)) {
		/* Ignore delayed exceptions from user space */
		asm volatile("1: fwait\n"
			     "2:\n"
			     _ASM_EXTABLE(1b, 2b));
		kgr__thread_fpu_end(tsk);
	}
}

static inline void kgr_drop_fpu(struct task_struct *tsk)
{
	/*
	 * Forget coprocessor state..
	 */
	preempt_disable();
	tsk->fpu_counter = 0;
	kgr__drop_fpu(tsk);
	clear_stopped_child_used_math(tsk);
	preempt_enable();
}

static inline void kgr_user_fpu_begin(void)
{
	preempt_disable();
	if (!user_has_fpu())
		kgr__thread_fpu_begin(current);
	preempt_enable();
}


/* from arch/x86/kernel/signal.c */
/* inlined */
static unsigned long kgr_align_sigframe(unsigned long sp)
{
#ifdef CONFIG_X86_32
	/*
	 * Align the stack pointer according to the i386 ABI,
	 * i.e. so that on function entry ((sp + 4) & 15) == 0.
	 */
	sp = ((sp + 4) & -16ul) - 4;
#else /* !CONFIG_X86_32 */
	sp = round_down(sp, 16) - 8;
#endif
	return sp;
}


/* inlined */
static inline void __user *
kgr_get_sigframe(struct k_sigaction *ka, struct pt_regs *regs,
		size_t frame_size, void __user **fpstate)
{
	/* Default to using normal stack */
	unsigned long math_size = 0;
	unsigned long sp = regs->sp;
	unsigned long buf_fx = 0;
	int onsigstack = on_sig_stack(sp);

	/* redzone */
	if (config_enabled(CONFIG_X86_64))
		sp -= 128;

	if (!onsigstack) {
		/* This is the X/Open sanctioned signal stack switching.  */
		if (ka->sa.sa_flags & SA_ONSTACK) {
			if (current->sas_ss_size)
				sp = current->sas_ss_sp + current->sas_ss_size;
		} else if (config_enabled(CONFIG_X86_32) &&
			   (regs->ss & 0xffff) != __USER_DS &&
			   !(ka->sa.sa_flags & SA_RESTORER) &&
			   ka->sa.sa_restorer) {
				/* This is the legacy signal stack switching. */
				sp = (unsigned long) ka->sa.sa_restorer;
		}
	}

	if (used_math()) {
		sp = alloc_mathframe(sp, config_enabled(CONFIG_X86_32),
				     &buf_fx, &math_size);
		*fpstate = (void __user *)sp;
	}

	sp = kgr_align_sigframe(sp - frame_size);

	/*
	 * If we are on the alternate signal stack and would overflow it, don't.
	 * Return an always-bogus address instead so we will die with SIGSEGV.
	 */
	if (onsigstack && !likely(on_sig_stack(sp)))
		return (void __user *)-1L;

	/* save i387 and extended state */
	if (used_math() &&
	    kgr_save_xstate_sig(*fpstate, (void __user *)buf_fx, math_size) < 0)
		return (void __user *)-1L;

	return (void __user *)sp;
}

/* inlined */
static int kgr__setup_rt_frame(int sig, struct ksignal *ksig,
			       sigset_t *set, struct pt_regs *regs)
{
	struct rt_sigframe __user *frame;
	void __user *fp = NULL;
	int err = 0;

	frame = kgr_get_sigframe(&ksig->ka, regs, sizeof(struct rt_sigframe),
				 &fp);

	if (!access_ok(VERIFY_WRITE, frame, sizeof(*frame)))
		return -EFAULT;

	if (ksig->ka.sa.sa_flags & SA_SIGINFO) {
		if (kgr_copy_siginfo_to_user(&frame->info, &ksig->info))
			return -EFAULT;
	}

	put_user_try {
		/* Create the ucontext.  */
		if (cpu_has_xsave)
			put_user_ex(UC_FP_XSTATE, &frame->uc.uc_flags);
		else
			put_user_ex(0, &frame->uc.uc_flags);
		put_user_ex(0, &frame->uc.uc_link);
		save_altstack_ex(&frame->uc.uc_stack, regs->sp);

		/* Set up to return from userspace.  If provided, use a stub
		   already in userspace.  */
		/* x86-64 should always use SA_RESTORER. */
		if (ksig->ka.sa.sa_flags & SA_RESTORER) {
			put_user_ex(ksig->ka.sa.sa_restorer, &frame->pretcode);
		} else {
			/* could use a vstub here */
			err |= -EFAULT;
		}
	} put_user_catch(err);

	err |= kgr_setup_sigcontext(&frame->uc.uc_mcontext, fp, regs, set->sig[0]);
	err |= __copy_to_user(&frame->uc.uc_sigmask, set, sizeof(*set));

	if (err)
		return -EFAULT;

	/* Set up registers for signal handler */
	regs->di = sig;
	/* In case the signal handler was declared without prototypes */
	regs->ax = 0;

	/* This also works for non SA_SIGINFO handlers because they expect the
	   next argument after the signal number on the stack. */
	regs->si = (unsigned long)&frame->info;
	regs->dx = (unsigned long)&frame->uc;
	regs->ip = (unsigned long) ksig->ka.sa.sa_handler;

	regs->sp = (unsigned long)frame;

	/* Set up the CS register to run signal handlers in 64-bit mode,
	   even if the handler happens to be interrupting 32-bit code. */
	regs->cs = __USER_CS;

	return 0;
}

/* inlined */
static int kgr_x32_setup_rt_frame(struct ksignal *ksig,
				  compat_sigset_t *set,
				  struct pt_regs *regs)
{
#ifdef CONFIG_X86_X32_ABI
	struct rt_sigframe_x32 __user *frame;
	void __user *restorer;
	int err = 0;
	void __user *fpstate = NULL;

	frame = get_sigframe(&ksig->ka, regs, sizeof(*frame), &fpstate);

	if (!access_ok(VERIFY_WRITE, frame, sizeof(*frame)))
		return -EFAULT;

	if (ksig->ka.sa.sa_flags & SA_SIGINFO) {
		if (copy_siginfo_to_user32(&frame->info, &ksig->info))
			return -EFAULT;
	}

	put_user_try {
		/* Create the ucontext.  */
		if (cpu_has_xsave)
			put_user_ex(UC_FP_XSTATE, &frame->uc.uc_flags);
		else
			put_user_ex(0, &frame->uc.uc_flags);
		put_user_ex(0, &frame->uc.uc_link);
		compat_save_altstack_ex(&frame->uc.uc_stack, regs->sp);
		put_user_ex(0, &frame->uc.uc__pad0);

		if (ksig->ka.sa.sa_flags & SA_RESTORER) {
			restorer = ksig->ka.sa.sa_restorer;
		} else {
			/* could use a vstub here */
			restorer = NULL;
			err |= -EFAULT;
		}
		put_user_ex(restorer, &frame->pretcode);
	} put_user_catch(err);

	err |= setup_sigcontext(&frame->uc.uc_mcontext, fpstate,
				regs, set->sig[0]);
	err |= __copy_to_user(&frame->uc.uc_sigmask, set, sizeof(*set));

	if (err)
		return -EFAULT;

	/* Set up registers for signal handler */
	regs->sp = (unsigned long) frame;
	regs->ip = (unsigned long) ksig->ka.sa.sa_handler;

	/* We use the x32 calling convention here... */
	regs->di = ksig->sig;
	regs->si = (unsigned long) &frame->info;
	regs->dx = (unsigned long) &frame->uc;

	loadsegment(ds, __USER_DS);
	loadsegment(es, __USER_DS);

	regs->cs = __USER_CS;
	regs->ss = __USER_DS;
#endif	/* CONFIG_X86_X32_ABI */

	return 0;
}

/* inlined */
static int kgr_signr_convert(int sig)
{
#ifdef CONFIG_X86_32
	struct thread_info *info = current_thread_info();

	if (info->exec_domain && info->exec_domain->signal_invmap && sig < 32)
		return info->exec_domain->signal_invmap[sig];
#endif /* CONFIG_X86_32 */
	return sig;
}

/* inlined */
static int
kgr_setup_rt_frame(struct ksignal *ksig, struct pt_regs *regs)
{
	int usig = kgr_signr_convert(ksig->sig);
	sigset_t *set = sigmask_to_save();
	compat_sigset_t *cset = (compat_sigset_t *) set;

	/* Set up the stack frame */
	if (is_ia32_frame()) {
		if (ksig->ka.sa.sa_flags & SA_SIGINFO)
			return kgr_ia32_setup_rt_frame(usig, ksig, cset, regs);
		else
			return kgr_ia32_setup_frame(usig, ksig, cset, regs);
	} else if (is_x32_frame()) {
		return kgr_x32_setup_rt_frame(ksig, cset, regs);
	} else {
		return kgr__setup_rt_frame(ksig->sig, ksig, set, regs);
	}
}

/* inlined */
static inline unsigned long kgr_get_nr_restart_syscall(const struct pt_regs *regs)
{
#if defined(CONFIG_X86_32) || !defined(CONFIG_X86_64)
	return __NR_restart_syscall;
#else /* !CONFIG_X86_32 && CONFIG_X86_64 */
	return test_thread_flag(TIF_IA32) ? __NR_ia32_restart_syscall :
		__NR_restart_syscall | (regs->orig_ax & __X32_SYSCALL_BIT);
#endif /* CONFIG_X86_32 || !CONFIG_X86_64 */
}


/* from include/linux/signal.h */
/* call the kallsyms-resolved kgr_signal_to_deliver() */
#define kgr_get_signal(ksig)					\
({								\
	struct ksignal *p = (ksig);				\
	p->sig = kgr_get_signal_to_deliver(&p->info, &p->ka,	\
					   signal_pt_regs(), NULL);	\
	p->sig > 0;						\
})


/* from include/linux/sched.h */
/* call the kallsyms-resolved kgr__set_current_blocked() */
static inline void kgr_restore_saved_sigmask(void)
{
	if (test_and_clear_restore_sigmask())
		kgr__set_current_blocked(&current->saved_sigmask);
}


/* from arch/x86/kernel/xsave.c */
/* inlined */
static inline int kgr_check_for_xstate(struct i387_fxsave_struct __user *buf,
				       void __user *fpstate,
				       struct _fpx_sw_bytes *fx_sw)
{
	int min_xstate_size = sizeof(struct i387_fxsave_struct) +
			      sizeof(struct xsave_hdr_struct);
	unsigned int magic2;

	if (__copy_from_user(fx_sw, &buf->sw_reserved[0], sizeof(*fx_sw)))
		return -1;

	/* Check for the first magic field and other error scenarios. */
	if (fx_sw->magic1 != FP_XSTATE_MAGIC1 ||
	    fx_sw->xstate_size < min_xstate_size ||
	    fx_sw->xstate_size > xstate_size ||
	    fx_sw->xstate_size > fx_sw->extended_size)
		return -1;

	/*
	 * Check for the presence of second magic word at the end of memory
	 * layout. This detects the case where the user just copied the legacy
	 * fpstate layout with out copying the extended state information
	 * in the memory layout.
	 */
	if (__get_user(magic2, (__u32 __user *)(fpstate + fx_sw->xstate_size))
	    || magic2 != FP_XSTATE_MAGIC2)
		return -1;

	return 0;
}

/* inlined */
static inline void
kgr_sanitize_restored_xstate(struct task_struct *tsk,
			     struct user_i387_ia32_struct *ia32_env,
			     u64 xstate_bv, int fx_only)
{
	struct xsave_struct *xsave = &tsk->thread.fpu.state->xsave;
	struct xsave_hdr_struct *xsave_hdr = &xsave->xsave_hdr;

	if (use_xsave()) {
		/* These bits must be zero. */
		xsave_hdr->reserved1[0] = xsave_hdr->reserved1[1] = 0;

		/*
		 * Init the state that is not present in the memory
		 * layout and not enabled by the OS.
		 */
		if (fx_only)
			xsave_hdr->xstate_bv = XSTATE_FPSSE;
		else
			xsave_hdr->xstate_bv &= (*kgr_pcntxt_mask & xstate_bv);
	}

	if (use_fxsr()) {
		/*
		 * mscsr reserved bits must be masked to zero for security
		 * reasons.
		 */
		xsave->i387.mxcsr &= *kgr_mxcsr_feature_mask;

		kgr_convert_to_fxsr(tsk, ia32_env);
	}
}

/* inlined */
static inline int kgr_restore_user_xstate(void __user *buf, u64 xbv,
					  int fx_only)
{
	if (use_xsave()) {
		if ((unsigned long)buf % 64 || fx_only) {
			u64 init_bv = *kgr_pcntxt_mask & ~XSTATE_FPSSE;
			xrstor_state(*kgr_init_xstate_buf, init_bv);
			return fxrstor_user(buf);
		} else {
			u64 init_bv = *kgr_pcntxt_mask & ~xbv;
			if (unlikely(init_bv))
				xrstor_state(*kgr_init_xstate_buf, init_bv);
			return xrestore_user(buf, xbv);
		}
	} else if (use_fxsr()) {
		return fxrstor_user(buf);
	} else
		return frstor_user(buf);
}


/* from arch/x86/kernel/i387.c */
/* inlined */
static inline bool kgr_interrupted_user_mode(void)
{
	struct pt_regs *regs = get_irq_regs();
	return regs && user_mode_vm(regs);
}



/*
 * patched
 *
 * Note that CONFIG_XEN brings its own definition, but it's identical.
 */
int kgr_arch_dup_task_struct(struct task_struct *dst, struct task_struct *src)
{
	int ret;

	*dst = *src;
	if (fpu_allocated(&src->thread.fpu)) {
		memset(&dst->thread.fpu, 0, sizeof(dst->thread.fpu));
		ret = fpu_alloc(&dst->thread.fpu);
		if (ret)
			return ret;
		fpu_copy(dst, src);
		/*
		 * Fix CVE-2018-3665
		 *  +5 lines
		 *
		 * Note that this hunk is not required from a security
		 * perspective. The fpu_copy() from above can call
		 * unlazy_fpu() which makes the current task not own
		 * the FP regs anymore. If current does some math
		 * inbetween the return from fork() and getting scheduled out,
		 * a device_not_available exception will be issued and
		 * the testcase ftracing that will fail. Reload current's FPU
		 * registers to prevent that from happening.
		 */
		if (!use_eager_fpu() && src == current && tsk_used_math(src)) {
			preempt_disable();
			kgr_math_state_restore();
			preempt_enable();
		}
	/*
	 * Fix CVE-2018-3665
	 *  -1 line, +8 lines
	 */
	} else if (!(dst->flags & PF_KTHREAD)) {
		/* Apparently not needed, but better be safe. */
		memset(&dst->thread.fpu, 0, sizeof(dst->thread.fpu));

		ret = init_fpu(dst);
		if (ret)
			return ret;
	}
	return 0;
}

/* patched, inlined */
static inline void kgr_drop_init_fpu(struct task_struct *tsk)
{
	if (!use_eager_fpu()) {
		/*
		 * Fix CVE-2018-3665
		 *  +2 lines
		 */
		preempt_disable();

		kgr_drop_fpu(tsk);

		/*
		 * Fix CVE-2018-3665
		 *  +28 lines
		 */
		if (fpu_allocated(&tsk->thread.fpu)) {
			/*
			 * This is basically init_fpu() for
			 * !tsk_used_math() (as cleared by the
			 * drop_fpu() from above) and fpu_allocated().
			 */
			fpu_finit(&tsk->thread.fpu);
			set_stopped_child_used_math(tsk);

			/*
			 * C.f. upstream commit 4ecd16ec7059 ("x86/fpu: Fix
			 * math emulation in eager fpu mode"). Note that we
			 * should always have X86_FEATURE_FPU on
			 * CONFIG_X86_64, but be defensive.
			 */
			if (static_cpu_has(X86_FEATURE_FPU)) {
				kgr__thread_fpu_begin(tsk);
				if (restore_fpu_checking(tsk)) {
					/*
					 * Something's really fishy. Call
					 * drop_fpu() once again and hope for
					 * the best.
					 */
					kgr_drop_fpu(tsk);
				}
			}
		}
		preempt_enable();

	} else {
		if (use_xsave())
			xrstor_state(*kgr_init_xstate_buf, -1);
		else
			fxrstor_checking(&(*kgr_init_xstate_buf)->i387);
	}
}

/*
 * patched
 *
 * Note that CONFIG_XEN brings its own definition, but it's identical.
 */
void kgr_flush_thread(void)
{
	struct task_struct *tsk = current;

	kgr_flush_ptrace_hw_breakpoint(tsk);
	memset(tsk->thread.tls_array, 0, sizeof(tsk->thread.tls_array));
	kgr_drop_init_fpu(tsk);
	/*
	 * Fix CVE-2018-3665
	 *  -6 lines
	 */
}

/* patched, inlined, calls inlined drop_init_fpu() */
static void
kgr_handle_signal(struct ksignal *ksig, struct pt_regs *regs)
{
	bool failed;
	/* Are we from a system call? */
	if (syscall_get_nr(current, regs) >= 0) {
		/* If so, check system call restarting.. */
		switch (syscall_get_error(current, regs)) {
		case -ERESTART_RESTARTBLOCK:
		case -ERESTARTNOHAND:
			regs->ax = -EINTR;
			break;

		case -ERESTARTSYS:
			if (!(ksig->ka.sa.sa_flags & SA_RESTART)) {
				regs->ax = -EINTR;
				break;
			}
		/* fallthrough */
		case -ERESTARTNOINTR:
			regs->ax = regs->orig_ax;
			regs->ip -= 2;
			break;
		}
	}

	/*
	 * If TF is set due to a debugger (TIF_FORCED_TF), clear the TF
	 * flag so that register information in the sigcontext is correct.
	 */
	if (unlikely(regs->flags & X86_EFLAGS_TF) &&
	    likely(test_and_clear_thread_flag(TIF_FORCED_TF)))
		regs->flags &= ~X86_EFLAGS_TF;

	failed = (kgr_setup_rt_frame(ksig, regs) < 0);
	if (!failed) {
		/*
		 * Clear the direction flag as per the ABI for function entry.
		 *
		 * Clear RF when entering the signal handler, because
		 * it might disable possible debug exception from the
		 * signal handler.
		 *
		 * Clear TF when entering the signal handler, but
		 * notify any tracer that was single-stepping it.
		 * The tracer may want to single-step inside the
		 * handler too.
		 */
		regs->flags &= ~(X86_EFLAGS_DF|X86_EFLAGS_RF|X86_EFLAGS_TF);
		/*
		 * Ensure the signal handler starts with the new fpu state.
		 */
		if (used_math())
			kgr_drop_init_fpu(current);
	}
	kgr_signal_setup_done(failed, ksig, test_thread_flag(TIF_SINGLESTEP));
}

/* patched, calls inlined handle_signal() */
void kgr_do_signal(struct pt_regs *regs)
{
	struct ksignal ksig;

	if (kgr_get_signal(&ksig)) {
		/* Whee! Actually deliver the signal.  */
		kgr_handle_signal(&ksig, regs);
		return;
	}

	/* Did we come from a system call? */
	if (syscall_get_nr(current, regs) >= 0) {
		/* Restart the system call - no handlers present */
		switch (syscall_get_error(current, regs)) {
		case -ERESTARTNOHAND:
		case -ERESTARTSYS:
		case -ERESTARTNOINTR:
			regs->ax = regs->orig_ax;
			regs->ip -= 2;
			break;

		case -ERESTART_RESTARTBLOCK:
			regs->ax = kgr_get_nr_restart_syscall(regs);
			regs->ip -= 2;
			break;
		}
	}

	/*
	 * If there's no signal to deliver, we just put the saved sigmask
	 * back.
	 */
	kgr_restore_saved_sigmask();
}

/* patched */
int kgr__restore_xstate_sig(void __user *buf, void __user *buf_fx, int size)
{
	int ia32_fxstate = (buf != buf_fx);
	struct task_struct *tsk = current;
	int state_size = xstate_size;
	u64 xstate_bv = 0;
	int fx_only = 0;

	ia32_fxstate &= (config_enabled(CONFIG_X86_32) ||
			 config_enabled(CONFIG_IA32_EMULATION));

	if (!buf) {
		kgr_drop_init_fpu(tsk);
		return 0;
	}

	if (!access_ok(VERIFY_READ, buf, size))
		return -EACCES;

	if (!used_math() && init_fpu(tsk))
		return -1;

	if (!static_cpu_has(X86_FEATURE_FPU))
		return fpregs_soft_set(current, NULL,
				       0, sizeof(struct user_i387_ia32_struct),
				       NULL, buf) != 0;

	if (use_xsave()) {
		struct _fpx_sw_bytes fx_sw_user;
		if (unlikely(kgr_check_for_xstate(buf_fx, buf_fx, &fx_sw_user))) {
			/*
			 * Couldn't find the extended state information in the
			 * memory layout. Restore just the FP/SSE and init all
			 * the other extended state.
			 */
			state_size = sizeof(struct i387_fxsave_struct);
			fx_only = 1;
		} else {
			state_size = fx_sw_user.xstate_size;
			xstate_bv = fx_sw_user.xstate_bv;
		}
	}

	if (ia32_fxstate) {
		/*
		 * For 32-bit frames with fxstate, copy the user state to the
		 * thread's fpu state, reconstruct fxstate from the fsave
		 * header. Sanitize the copied state etc.
		 */
		struct fpu *fpu = &tsk->thread.fpu;
		struct user_i387_ia32_struct env;
		int err = 0;

		/*
		 * Drop the current fpu which clears used_math(). This ensures
		 * that any context-switch during the copy of the new state,
		 * avoids the intermediate state from getting restored/saved.
		 * Thus avoiding the new restored state from getting corrupted.
		 * We will be ready to restore/save the state only after
		 * set_used_math() is again set.
		 */
		kgr_drop_fpu(tsk);

		if (__copy_from_user(&fpu->state->xsave, buf_fx, state_size) ||
		    __copy_from_user(&env, buf, sizeof(env))) {
			fpu_finit(fpu);
			err = -1;
		} else {
			kgr_sanitize_restored_xstate(tsk, &env, xstate_bv, fx_only);
		}

		set_used_math();
		/*
		 * Fix CVE-2018-3665
		 *  -1 line, +6 lines
		 */
		/*
		 * C.f. upstream commit 4ecd16ec7059 ("x86/fpu: Fix math
		 * emulation in eager fpu mode"). Note that we should always
		 * have X86_FEATURE_FPU on CONFIG_X86_64, but be defensive.
		 */
		if (use_eager_fpu() || static_cpu_has(X86_FEATURE_FPU)) {
			preempt_disable();
			kgr_math_state_restore();
			preempt_enable();
		}

		return err;
	} else {
		/*
		 * For 64-bit frames and 32-bit fsave frames, restore the user
		 * state to the registers directly (with exceptions handled).
		 */
		kgr_user_fpu_begin();
		if (kgr_restore_user_xstate(buf_fx, xstate_bv, fx_only)) {
			kgr_drop_init_fpu(tsk);
			return -1;
		}
	}

	return 0;
}

/* patched */
void kgr__kernel_fpu_end(void)
{
	if (use_eager_fpu()) {
		/*
		 * For eager fpu, most the time, tsk_used_math() is true.
		 * Restore the user math as we are done with the kernel usage.
		 * At few instances during thread exit, signal handling etc,
		 * tsk_used_math() is false. Those few places will take proper
		 * actions, so we don't need to restore the math here.
		 */
		if (likely(tsk_used_math(current)))
			kgr_math_state_restore();
	} else {
		/*
		 * Fix CVE-2018-3665
		 *  -1 line, +27 lines
		 */
		if (in_interrupt() && !kgr_interrupted_user_mode()) {
			/*
			 * Careful, careful here. The
			 * kernel_fpu_begin() was protected by a
			 * irq_fpu_usable() from driver code, which
			 * succeeded. In particular,
			 * interrupted_kernel_fpu_idle() was true.
			 * Another kernel_fpu_begin() might follow and
			 * there, interrupted_kernel_fpu_idle() must
			 * also evaluate to true.
			 *
			 * Note that we don't have to wipe out the FPU regs,
			 * because there are two possibilities:
			 * 1. we interrupted a kernel thread or
			 * 2. we interrupted a user thread while in kernel mode.
			 *
			 * For 1., there'll be a context switch before
			 * userspace can access the FPU again. For
			 * 2., the eager switch logic almost always
			 * enforces the current task to own the FPU,
			 * and thus, interrupted_kernel_fpu_idle()
			 * would have failed in the first place.
			 */
			stts();

		} else if (likely(tsk_used_math(current))) {
			kgr_math_state_restore();

		} else {
			stts();

		}
	}
}



/* based on arch/x86/include/asm/fpu-internal.h */
/*
 * Like fpu_alloc(), but with GFP_NOWAIT for use in
 * trace_sched_switch()'s atomic context.
 */
static int kgr_fpu_alloc_nowait(struct fpu *fpu)
{
	if (fpu_allocated(fpu))
		return 0;
	fpu->state = kmem_cache_alloc(task_xstate_cachep, GFP_NOWAIT);
	if (!fpu->state)
		return -ENOMEM;
	WARN_ON((unsigned long)fpu->state & 15);
	return 0;
}

/*
 * Like init_fpu(), but without the !!tsk_used_math() branch
 * and with replacing the GFP_KERNEL fpu_alloc() by
 * the GFP_NOWAIT kgr_fpu_alloc_nowait().
 */
static int kgr_init_fpu_nowait(struct task_struct *tsk)
{
	int ret;

	/*
	 * Memory allocation at the first usage of the FPU and other state.
	 */
	ret = kgr_fpu_alloc_nowait(&tsk->thread.fpu);
	if (ret)
		return ret;

	fpu_finit(&tsk->thread.fpu);

	set_stopped_child_used_math(tsk);
	return 0;
}


/*
 * This hook will be installed at the sched_switch tracepoint. It
 * tricks the subsequently called switch_fpu_prepare() into eager
 * switching behaviour.
 */
static void trick_eager_fpu_sched_switch(void *data,
					struct task_struct *prev,
					struct task_struct *next)
{
	/*
	 * Trick the lazy FPU switching heuristic in
	 * switch_fpu_prepare() into !!use_eager_fpu() behaviour.
	 */
	if (!tsk_used_math(next) && !(next->flags & PF_KTHREAD)) {
		if (task_thread_info(next)->status & TS_COMPAT) {
			/*
			 * This task might have been scheduled out
			 * while in sys32_sigreturn or
			 * sys32_rt_sigreturn. More precisely, in
			 * __restore_xstate_sig()'s !!ia32_fxstate path
			 * where we must not concurrently reinit the FPU
			 * state here.
			 */
			return;
		}
		/*
		 * The next task doesn't have its FPU state
		 * initialized and thus, switch_fpu_prepare()/_finish()
		 * won't update the FP regs.
		 *
		 * Note that with eagerfpu, threads always have their
		 * FPU state initialized after execve()ing because of
		 * flush_thread()->drop_init_fpu().
		 *
		 * Deliberately ignore any memory allocation failure
		 * here.
		 */
		kgr_init_fpu_nowait(next);
	}

	if (tsk_used_math(next)) {
		/* Touch ->fpu_counter only if needed. */
		next->fpu_counter = next->fpu_counter > 5 ? next->fpu_counter : 6;
	}
}



static bool sched_switch_probe_registered = false;

static void bsc1090338_activate(struct kgr_subpatch_data *header)
{
	int ret;

	if (use_eager_fpu()) {
		/* Everything's alright already, do nothing. */
		return;
	}

	/*
	 * C.f. upstream commit 4ecd16ec7059 ("x86/fpu: Fix math
	 * emulation in eager fpu mode"). Note that we should always
	 * have X86_FEATURE_FPU on CONFIG_X86_64, but be defensive.
	 */
	if (!static_cpu_has(X86_FEATURE_FPU))
		return;

	ret = register_trace_sched_switch(trick_eager_fpu_sched_switch, NULL);
	if (ret) {
		pr_err("failed to register sched switch probe: %d\n", ret);
		return;
	}
	sched_switch_probe_registered = true;
}

static void bsc1090338_deactivate(struct kgr_subpatch_data *header)
{
	int ret;

	if (!sched_switch_probe_registered)
		return;

	ret = unregister_trace_sched_switch(trick_eager_fpu_sched_switch, NULL);
	if (ret) {
		/*
		 * That's impossible, but for debugging purposes,
		 * print an error.
		 */
		pr_err("failed to unregister sched switch probe: %d\n", ret);
	}
}

static bool bsc1090338_prepare_migration(struct kgr_subpatch *self,
					 struct kgr_subpatch_data *prev_data)
{
	return true;
}

static struct kgr_subpatch bsc1090338_subpatch = {
	.mod = THIS_MODULE,
	.alloc_data = NULL, /* There is no shared state. */
	.free_data = NULL,

	.prepare_migration = bsc1090338_prepare_migration,

	.post_patch = bsc1090338_activate,
	.pre_migrate_to = bsc1090338_activate,
	.pre_revert = bsc1090338_deactivate,
	.post_migrate_away = bsc1090338_deactivate,
};


static int kgr_patch_bsc1090338_kallsyms(void)
{
	unsigned long addr;
	int i;

	for (i = 0; i < ARRAY_SIZE(kgr_funcs); i++) {
		addr = kallsyms_lookup_name(kgr_funcs[i].name);
		if (!addr) {
			pr_err("symbol %s not resolved\n", kgr_funcs[i].name);
			return -ENOENT;
		}

		*(kgr_funcs[i].addr) = (void *)addr;
	}

	return 0;
}

int kgr_patch_bsc1090338_init(void)
{
	int ret;

	ret =  kgr_patch_bsc1090338_kallsyms();
	if (ret)
		return ret;

	ret = kgr_subpatch_register(BSC1090338_SUBPATCH_ID,
				    &bsc1090338_subpatch);
	return ret;
}

void kgr_patch_bsc1090338_cleanup(void)
{
	kgr_subpatch_unregister(&bsc1090338_subpatch);

	if (sched_switch_probe_registered)
		kgr_tracepoint_synchronize_unregister();
}

#endif /* CONFIG_X86_64 */
